import 'dart:async';
import 'dart:io';
import 'dart:typed_data';
import 'package:flutter/services.dart';
import 'package:flutter_ble_lib/flutter_ble_lib.dart';
import 'package:permission_handler/permission_handler.dart';

import 'package:hiflex/hiflex_device.dart';
import 'package:hiflex/hiflex_raw_device.dart';

class HiflexManager {
  static final String btFtsameService = "0000180a-0000-1000-8000-00805f9b34fb";

  static const MethodChannel _channel = const MethodChannel('hiflex');

  static Future<String> get platformVersion async {
    final String version = await _channel.invokeMethod('getPlatformVersion');
    return version;
  }

  static final HiflexManager _instance = new HiflexManager._internal();

  factory HiflexManager() {
    return _instance;
  }

  HiflexManager._internal();

  BleManager _manager = BleManager();
  List<HiflexRawDevice> _deviceList = [];

  Future<bool> init() async {
    checkPermissions() async {
      if (Platform.isAndroid) {
        if (await Permission.contacts.request().isGranted) {}
        Map<Permission, PermissionStatus> statuses =
            await [Permission.location].request();
        print("Location Permission: ${statuses[Permission.location]}");
      }
    }

    if (Platform.isAndroid) {
      await _manager
          .createClient(
            restoreStateIdentifier: "hiflex.restore.state.identifier",
            restoreStateAction: (peripherals) {
              peripherals?.forEach((peripheral) {
                print("Restored peripheral: ${peripheral.name}");
              });
            },
          )
          .catchError((e) => print("Couldn't create BLE client  $e"))
          .then((_) => checkPermissions())
          .catchError((e) => print("Permission check error $e"));
    } else if (Platform.isIOS) {
      await _manager.createClient();
    } else
      return false;

    return true;
  }

  Future<HiflexDevice> findDevice() async {
    var completer = new Completer<HiflexDevice>();
    if (await _manager.bluetoothState() == BluetoothState.POWERED_ON) {
      _manager.startPeripheralScan().listen((scanResult) async {
        var findDevice = _deviceList.firstWhere(
            (element) => element.identifier == scanResult.peripheral.identifier,
            orElse: () => null);

        if (findDevice == null) {
          var rawDevice = new HiflexRawDevice(scanResult);
          _deviceList.add(rawDevice);

          if (rawDevice.existsAdService(btFtsameService)) {
            _manager.stopPeripheralScan();
            var newDevice = await fectoryHiflexDevice(rawDevice);
            if (newDevice != null && await newDevice.connect()) {
              completer.complete(newDevice);
            }
          }
        } else
          findDevice.update(scanResult);
      });
      return completer.future;
    }
    return null;
  }

  Future<HiflexDevice> fectoryHiflexDevice(HiflexRawDevice rawDevice) async {
    Future<Service> getFtsameService() async {
      var uuid = HiflexDeviceBT.btFtsameService;
      var services = await rawDevice.peripheral.services();
      if (services != null) {
        return services.firstWhere((element) => element.uuid == uuid);
      }
      return null;
    }

    Future<String> getCharValue(Service service, String uuid) async {
      var characteristic = await service.readCharacteristic(uuid);
      if (characteristic != null) {
        return new String.fromCharCodes(characteristic.value);
      }
      return "";
    }

    final btModelNumber = HiflexDeviceBT.btModelNumber;
    final btHardwareRevision = HiflexDeviceBT.btHardwareRevision;
    final btFirmwareRevision = HiflexDeviceBT.btFirmwareRevision;

    await rawDevice.peripheral.connect();
    if (await rawDevice.peripheral.isConnected()) {
      await rawDevice.peripheral.discoverAllServicesAndCharacteristics();

      var service = await getFtsameService();
      if (service != null) {
        var modelNumber = await getCharValue(service, btModelNumber);
        var hardware = await getCharValue(service, btHardwareRevision);
        var firmware = await getCharValue(service, btFirmwareRevision);

        rawDevice.peripheral.observeConnectionState().listen((event) {
          if (event == PeripheralConnectionState.disconnected) {
            _deviceList.remove(rawDevice);
          }
        });

        return new HiflexDeviceBT(rawDevice, modelNumber, hardware, firmware);
      }
    }

    return null;
  }

  printAdvertisementData(String name, AdvertisementData ad) {
    if (ad.serviceUuids != null) {
      print("$name Service UUIDs");
      for (var uuid in ad.serviceUuids) {
        print(" - $uuid");
      }
    }

    if (ad.serviceData != null) {
      print("$name Service Data");
      for (var key in ad.serviceData.keys) {
        print(" - $key = ${byteToString(ad.serviceData[key])}");
      }
    }

    if (ad.solicitedServiceUuids != null) {
      print("$name Solicited Service UUIDs");
      for (var uuid in ad.solicitedServiceUuids) {
        print(" - $uuid");
      }
    }
  }

  String byteToString(Uint8List rawData) {
    var result = "[${rawData.lengthInBytes}] ";
    var bytes = ByteData.sublistView(rawData);
    for (int n = 0; n < bytes.lengthInBytes; ++n) {
      result += "${bytes.getUint8(n).toRadixString(16)}:";
    }
    return result;
  }
}
