import 'dart:typed_data';
import 'dart:math';

class Quaternion {
  double w;
  double x;
  double y;
  double z;

  Quaternion({this.w, this.x, this.y, this.z});

  static Quaternion get identity => Quaternion(w: 1, x: 0, y: 0, z: 0);
}

class EulurAngle {
  double x;
  double y;
  double z;

  EulurAngle({this.x, this.y, this.z});
}

class Vector3 {
  double x;
  double y;
  double z;

  Vector3({this.x, this.y, this.z});
}

class SamplingData {
  double stretch;
  Quaternion quaternion;
  EulurAngle eulurAngle;
  Vector3 acceleometer;
  Vector3 gyroscope;
  Vector3 magnetometer;
  double deltaTime;
}

class HiflexSampling {
  Map<String, HiflexScalar> _rawData = new Map();
  DateTime _privTime = new DateTime.now();
  Quaternion _quaternion = Quaternion.identity;

  HiflexSampling() {
    _rawData.addAll({
      "stretch": new HiflexScalarKalman(),
      "acc.x": new HiflexScalarKalman(),
      "acc.y": new HiflexScalarKalman(),
      "acc.z": new HiflexScalarKalman(),
      "gyr.x": new HiflexScalarKalman(),
      "gyr.y": new HiflexScalarKalman(),
      "gyr.z": new HiflexScalarKalman(),
      "mag.x": new HiflexScalarKalman(),
      "mag.y": new HiflexScalarKalman(),
      "mag.z": new HiflexScalarKalman(),
      "battery": new HiflexScalarNormal(),
      "deltaTime": new HiflexScalarKalman(),
    });
  }

  SamplingData get data {
    var sampling = new SamplingData();
    sampling.stretch = _rawData["stretch"].load();
    sampling.deltaTime = _rawData["deltaTime"].load();
    sampling.acceleometer = new Vector3(
      x: _rawData["acc.x"].load(),
      y: _rawData["acc.y"].load(),
      z: _rawData["acc.z"].load(),
    );
    sampling.gyroscope = new Vector3(
      x: _rawData['gyr.x'].load(),
      y: _rawData['gyr.y'].load(),
      z: _rawData['gyr.z'].load(),
    );
    sampling.magnetometer = new Vector3(
      x: _rawData['mag.x'].load(),
      y: _rawData['mag.y'].load(),
      z: _rawData['mag.z'].load(),
    );

    sampling.quaternion = new Quaternion(
      w: _quaternion.w,
      x: _quaternion.x,
      y: _quaternion.y,
      z: _quaternion.z,
    );
    sampling.eulurAngle = _toEulurAngle(_quaternion);

    return sampling;
  }

  int get batteryLevel => _rawData['battery'].load().toInt();

  ///
  SamplingData load(Uint8List rawData) {
    // Stream으로 전달.???
    var bytes = ByteData.sublistView(rawData);

    // 여기서 Quaternion 계산 후 _rawData에 저장.
    _rawData["acc.x"].store(bytes.getInt16(3, Endian.little).toDouble());
    _rawData["acc.y"].store(bytes.getInt16(5, Endian.little).toDouble());
    _rawData["acc.z"].store(bytes.getInt16(7, Endian.little).toDouble());
    _rawData["gyr.x"].store(bytes.getInt16(9, Endian.little).toDouble() / 65.6);
    _rawData["gyr.y"]
        .store(bytes.getInt16(11, Endian.little).toDouble() / 65.6);
    _rawData["gyr.z"]
        .store(bytes.getInt16(13, Endian.little).toDouble() / 65.6);
    _rawData["mag.x"].store(bytes.getInt16(15, Endian.little).toDouble());
    _rawData["mag.y"].store(bytes.getInt16(17, Endian.little).toDouble());
    _rawData["mag.z"].store(bytes.getInt16(19, Endian.little).toDouble());

    _madgwickAHRSupdate(
      _rawData["gyr.x"].raw * 0.0175,
      _rawData["gyr.y"].raw * 0.0175,
      _rawData["gyr.z"].raw * 0.0175,
      _rawData["acc.x"].raw,
      _rawData["acc.y"].raw,
      _rawData["acc.z"].raw,
      _rawData["mag.x"].raw,
      _rawData["mag.y"].raw,
      _rawData["mag.z"].raw,
    );

    // HiFLEX Softsensor 데이터 갱신.
    _rawData["stretch"].store(bytes.getUint16(21, Endian.big).toDouble());
    _rawData["battery"].store(bytes.getInt8(23).toDouble());

    // 수신 시간 기록.
    var nowTime = new DateTime.now();
    _rawData["deltaTime"]
        .store(nowTime.difference(_privTime).inMilliseconds.toDouble());
    _privTime = nowTime;

    return this.data;
  }

  String _byteToString(Uint8List rawData) {
    var result = "[${rawData.lengthInBytes}] ";
    var bytes = ByteData.sublistView(rawData);
    for (int n = 0; n < bytes.lengthInBytes; ++n) {
      result += "${bytes.getUint8(n).toRadixString(16)}:";
    }
    return result;
  }

  _madgwickAHRSupdate(double gx, double gy, double gz, double ax, double ay,
      double az, double mx, double my, double mz) {
    final sampleFreq = 142.857143; // sample frequency in Hz
    final betaDef = 0.1; // 2 * proportional gain (Kp)

    double recipNorm;
    double s0, s1, s2, s3;
    var q0 = _quaternion.w;
    var q1 = _quaternion.x;
    var q2 = _quaternion.y;
    var q3 = _quaternion.z;

    var qDot1 = 0.5 * (-q1 * gx - q2 * gy - q3 * gz);
    var qDot2 = 0.5 * (q0 * gx + q2 * gz - q3 * gy);
    var qDot3 = 0.5 * (q0 * gy - q1 * gz + q3 * gx);
    var qDot4 = 0.5 * (q0 * gz + q1 * gy - q2 * gx);

    if (!((ax == 0.0) && (ay == 0.0) && (az == 0.0))) {
      // Normalise accelerometer measurement
      recipNorm = _invSqrt(ax * ax + ay * ay + az * az);
      ax *= recipNorm;
      ay *= recipNorm;
      az *= recipNorm;

      // Normalise magnetometer measurement
      recipNorm = _invSqrt(mx * mx + my * my + mz * mz);
      mx *= recipNorm;
      my *= recipNorm;
      mz *= recipNorm;

      // Auxiliary variables to avoid repeated arithmetic
      var _2q0mx = 2.0 * q0 * mx;
      var _2q0my = 2.0 * q0 * my;
      var _2q0mz = 2.0 * q0 * mz;
      var _2q1mx = 2.0 * q1 * mx;
      var _2q0 = 2.0 * q0;
      var _2q1 = 2.0 * q1;
      var _2q2 = 2.0 * q2;
      var _2q3 = 2.0 * q3;
      var _2q0q2 = 2.0 * q0 * q2;
      var _2q2q3 = 2.0 * q2 * q3;
      var q0q0 = q0 * q0;
      var q0q1 = q0 * q1;
      var q0q2 = q0 * q2;
      var q0q3 = q0 * q3;
      var q1q1 = q1 * q1;
      var q1q2 = q1 * q2;
      var q1q3 = q1 * q3;
      var q2q2 = q2 * q2;
      var q2q3 = q2 * q3;
      var q3q3 = q3 * q3;

      // Reference direction of Earth's magnetic field
      var hx = mx * q0q0 -
          _2q0my * q3 +
          _2q0mz * q2 +
          mx * q1q1 +
          _2q1 * my * q2 +
          _2q1 * mz * q3 -
          mx * q2q2 -
          mx * q3q3;
      var hy = _2q0mx * q3 +
          my * q0q0 -
          _2q0mz * q1 +
          _2q1mx * q2 -
          my * q1q1 +
          my * q2q2 +
          _2q2 * mz * q3 -
          my * q3q3;
      var _2bx = sqrt(hx * hx + hy * hy);
      var _2bz = -_2q0mx * q2 +
          _2q0my * q1 +
          mz * q0q0 +
          _2q1mx * q3 -
          mz * q1q1 +
          _2q2 * my * q3 -
          mz * q2q2 +
          mz * q3q3;
      var _4bx = 2.0 * _2bx;
      var _4bz = 2.0 * _2bz;

      // Gradient decent algorithm corrective step
      s0 = -_2q2 * (2.0 * q1q3 - _2q0q2 - ax) +
          _2q1 * (2.0 * q0q1 + _2q2q3 - ay) -
          _2bz * q2 * (_2bx * (0.5 - q2q2 - q3q3) + _2bz * (q1q3 - q0q2) - mx) +
          (-_2bx * q3 + _2bz * q1) *
              (_2bx * (q1q2 - q0q3) + _2bz * (q0q1 + q2q3) - my) +
          _2bx * q2 * (_2bx * (q0q2 + q1q3) + _2bz * (0.5 - q1q1 - q2q2) - mz);
      s1 = _2q3 * (2.0 * q1q3 - _2q0q2 - ax) +
          _2q0 * (2.0 * q0q1 + _2q2q3 - ay) -
          4.0 * q1 * (1 - 2.0 * q1q1 - 2.0 * q2q2 - az) +
          _2bz * q3 * (_2bx * (0.5 - q2q2 - q3q3) + _2bz * (q1q3 - q0q2) - mx) +
          (_2bx * q2 + _2bz * q0) *
              (_2bx * (q1q2 - q0q3) + _2bz * (q0q1 + q2q3) - my) +
          (_2bx * q3 - _4bz * q1) *
              (_2bx * (q0q2 + q1q3) + _2bz * (0.5 - q1q1 - q2q2) - mz);
      s2 = -_2q0 * (2.0 * q1q3 - _2q0q2 - ax) +
          _2q3 * (2.0 * q0q1 + _2q2q3 - ay) -
          4.0 * q2 * (1 - 2.0 * q1q1 - 2.0 * q2q2 - az) +
          (-_4bx * q2 - _2bz * q0) *
              (_2bx * (0.5 - q2q2 - q3q3) + _2bz * (q1q3 - q0q2) - mx) +
          (_2bx * q1 + _2bz * q3) *
              (_2bx * (q1q2 - q0q3) + _2bz * (q0q1 + q2q3) - my) +
          (_2bx * q0 - _4bz * q2) *
              (_2bx * (q0q2 + q1q3) + _2bz * (0.5 - q1q1 - q2q2) - mz);
      s3 = _2q1 * (2.0 * q1q3 - _2q0q2 - ax) +
          _2q2 * (2.0 * q0q1 + _2q2q3 - ay) +
          (-_4bx * q3 + _2bz * q1) *
              (_2bx * (0.5 - q2q2 - q3q3) + _2bz * (q1q3 - q0q2) - mx) +
          (-_2bx * q0 + _2bz * q2) *
              (_2bx * (q1q2 - q0q3) + _2bz * (q0q1 + q2q3) - my) +
          _2bx * q1 * (_2bx * (q0q2 + q1q3) + _2bz * (0.5 - q1q1 - q2q2) - mz);
      recipNorm = _invSqrt(
          s0 * s0 + s1 * s1 + s2 * s2 + s3 * s3); // normalise step magnitude
      s0 *= recipNorm;
      s1 *= recipNorm;
      s2 *= recipNorm;
      s3 *= recipNorm;

      // Apply feedback step
      qDot1 -= betaDef * s0;
      qDot2 -= betaDef * s1;
      qDot3 -= betaDef * s2;
      qDot4 -= betaDef * s3;

      // Integrate rate of change of quaternion to yield quaternion
      q0 += qDot1 * (1.0 / sampleFreq);
      q1 += qDot2 * (1.0 / sampleFreq);
      q2 += qDot3 * (1.0 / sampleFreq);
      q3 += qDot4 * (1.0 / sampleFreq);

      recipNorm = _invSqrt(q0 * q0 + q1 * q1 + q2 * q2 + q3 * q3);
      q0 *= recipNorm;
      q1 *= recipNorm;
      q2 *= recipNorm;
      q3 *= recipNorm;

      _quaternion.w = q0;
      _quaternion.x = q1;
      _quaternion.y = q2;
      _quaternion.z = q3;
    }
  }

  double _invSqrt(double value) => 1.0 / sqrt(value);

  EulurAngle _toEulurAngle(Quaternion quaternion) {
    var q0 = quaternion.w;
    var q1 = quaternion.x;
    var q2 = quaternion.y;
    var q3 = quaternion.z;

    return new EulurAngle(
      x: asin(-2 * q1 * q3 + 2 * q0 * q2) * 57.3,
      y: atan2(2 * q2 * q3 + 2 * q0 * q1, -2 * q1 * q1 - 2 * q2 * q2 + 1) *
          57.3,
      z: atan2(-2 * q1 * q2 - 2 * q0 * q3, 2 * q2 * q2 + 2 * q3 * q3 - 1) *
          57.3,
    );
  }
}

/*
 * 칼만 필터를 적용해야 하는 부분이 있음.
 */
class HiflexScalar {
  double get raw => 0.0;
  double load() => 0.0;
  store(double changeValue) {}
}

abstract class HiflexScalarBase implements HiflexScalar {}

class HiflexScalarNormal extends HiflexScalarBase {
  double _value;

  double get raw => _value;
  double load() => _value;
  store(double changeValue) {
    _value = changeValue;
  }
}

class HiflexScalarKalman extends HiflexScalarBase {
  double _errMeasure = 2.0;
  double _errEstimate = 2.0;
  double _q = 0.01;
  double _currentEstimate = 0.0;
  double _lastEstimate = 0.0;
  double _kalmanGain = 0.0;
  double _rawValue = 0.0;

  double get raw => _rawValue;
  double load() => _currentEstimate;
  store(double changeValue) {
    _kalmanGain = _errEstimate / (_errEstimate + _errMeasure);
    _currentEstimate =
        _lastEstimate + _kalmanGain * (changeValue - _lastEstimate);
    _errEstimate = (1.0 - _kalmanGain) * _errEstimate +
        (_lastEstimate - _currentEstimate).abs() * _q;
    _lastEstimate = _currentEstimate;
    _rawValue = changeValue;
  }
}
